unit agg2d_main;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  ExtCtrls ,Agg2D, StdCtrls, ComCtrls, ImgList ;

const
 FLIP_Y = true;

var
 FONT_TIMES ,FONT_ARIAL ,FONT_VERDA : AnsiString;

type
  TForm1 = class(TForm)
    Image1: TImage;
    TrackBar1: TTrackBar;
    Image2: TImage;
    procedure FormDestroy(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure FormResize(Sender: TObject);
    procedure TrackBar1Change(Sender: TObject);
  private
    VG: TAgg2D;

  public
    { Public declarations }
  end;

var
  Form1: TForm1;

implementation

{$R *.DFM}

procedure TForm1.FormDestroy(Sender: TObject);
begin
 if Assigned(VG ) then
  VG.Free;

end;

procedure TForm1.FormCreate(Sender: TObject);
begin
 ControlStyle:=ControlStyle + [csOpaque ];

{ Vector Graphics Engine }
 VG:=TAgg2D.Create;

{ Drawing Surface }
 //Image1.Picture.Bitmap.PixelFormat:=pf24bit;
 Image1.Picture.Bitmap.PixelFormat:=pf32bit;

{ Source of Bitmap data }
 Image2.Picture.Bitmap.PixelFormat:=pf32bit;

// We must force Alpha Channel to be Opaque (255), because Delphi's conversion in previous line
// kept Alpha Channel Trasparent (0). You can set other values as well to achieve various levels
// of default TBitmap transparency
 BitmapAlphaTransparency(Image2.Picture.Bitmap ,255 );

end;

procedure TForm1.FormResize(Sender: TObject);
var
 xb1 ,yb1 ,xb2 ,yb2 : double;

 clr ,c1 ,c2 ,c3 : TAggColor;

 parl : array[0..5 ] of double;

begin
 Image1.Picture.Bitmap.Width :=Image1.Width;
 Image1.Picture.Bitmap.Height:=Image1.Height;

 if Assigned(VG ) then
  if VG.attach(Image1.Picture.Bitmap ,FLIP_Y ) then
   begin
   // Drawing Routine
    VG.ClearAll(255 ,255 ,255 );
    VG.AntiAliasGamma(1.4 );

    VG.FlipText (true );
    VG.ImageFlip(true );

   // Transfornations - Rotate around (300,300) to 5 degree
    VG.Translate(-300 ,-300 );
    VG.Rotate   (Deg2Rad(TrackBar1.Position ) );
    VG.Translate(300 ,300 );

   // Viewport - set 0,0,600,600 to the actual window size
   // preserving aspect ratio and placing the viewport in the center.
   // To ignore aspect ratio use Agg2D::Anisotropic
   // Note that the viewport just adds transformations to the current
   // affine matrix. So that, set the viewport *after* all transformations!
    VG.Viewport(
     0 ,0 ,600 ,600 ,
     0 ,0 ,Image1.Width ,Image1.Height ,
     //AGG_Anisotropic );
     AGG_XMidYMid );

   // Rounded Rect
    VG.LineColor(0 ,0 ,0 );
    VG.NoFill;
    VG.RoundedRect(0.5 ,0.5 ,600 - 0.5 ,600 - 0.5 ,20.0 );

   // Regular Text
    VG.Font(FONT_TIMES ,14.0 ,false ,false ,AGG_RasterFontCache );
    VG.FillColor(0 ,0 ,0 );
    VG.NoLine;
    VG.Text(100 ,20 ,'Regular Raster Text -- Fast, but can''t be rotated' );

   // Outlined Text
    VG.Font(FONT_TIMES ,50.0 ,false ,false ,AGG_VectorFontCache );
    VG.LineColor(50 ,0 ,0 );
    VG.FillColor(180 ,200 ,100 );
    VG.LineWidth(1.0 );
    VG.Text(100.5 ,50.5 ,'Outlined Text' );

   // Text Alignment
    VG.LineColor(0 ,0 ,0 );
    VG.Line(250.5 - 150 ,150.5      ,250.5 + 150 ,150.5      );
    VG.Line(250.5       ,150.5 - 20 ,250.5       ,150.5 + 20 );
    VG.Line(250.5 - 150 ,200.5      ,250.5 + 150 ,200.5      );
    VG.Line(250.5       ,200.5 - 20 ,250.5       ,200.5 + 20 );
    VG.Line(250.5 - 150 ,250.5      ,250.5 + 150 ,250.5      );
    VG.Line(250.5       ,250.5 - 20 ,250.5       ,250.5 + 20 );
    VG.Line(250.5 - 150 ,300.5      ,250.5 + 150 ,300.5      );
    VG.Line(250.5       ,300.5 - 20 ,250.5       ,300.5 + 20 );
    VG.Line(250.5 - 150 ,350.5      ,250.5 + 150 ,350.5      );
    VG.Line(250.5       ,350.5 - 20 ,250.5       ,350.5 + 20 );
    VG.Line(250.5 - 150 ,400.5      ,250.5 + 150 ,400.5      );
    VG.Line(250.5       ,400.5 - 20 ,250.5       ,400.5 + 20 );
    VG.Line(250.5 - 150 ,450.5      ,250.5 + 150 ,450.5      );
    VG.Line(250.5       ,450.5 - 20 ,250.5       ,450.5 + 20 );
    VG.Line(250.5 - 150 ,500.5      ,250.5 + 150 ,500.5      );
    VG.Line(250.5       ,500.5 - 20 ,250.5       ,500.5 + 20 );
    VG.Line(250.5 - 150 ,550.5      ,250.5 + 150 ,550.5      );
    VG.Line(250.5       ,550.5 - 20 ,250.5       ,550.5 + 20 );

    VG.FillColor(100 ,50 ,50 );
    VG.NoLine;
    VG.TextHints(false );
    VG.Font(FONT_TIMES ,40.0 ,false ,false ,AGG_VectorFontCache );

    VG.TextAlignment(AGG_AlignLeft ,AGG_AlignBottom );
    VG.Text(250.0 ,150.0 ,'Left-Bottom' ,true ,0 ,0 );

    VG.TextAlignment(AGG_AlignCenter ,AGG_AlignBottom );
    VG.Text(250.0 ,200.0 ,'Center-Bottom' ,true ,0 ,0 );

    VG.TextAlignment(AGG_AlignRight ,AGG_AlignBottom );
    VG.Text(250.0 ,250.0 ,'Right-Bottom' ,true ,0 ,0 );

    VG.TextAlignment(AGG_AlignLeft ,AGG_AlignCenter );
    VG.Text(250.0 ,300.0 ,'Left-Center' ,true ,0 ,0 );

    VG.TextAlignment(AGG_AlignCenter ,AGG_AlignCenter );
    VG.Text(250.0 ,350.0 ,'Center-Center' ,true ,0 ,0 );

    VG.TextAlignment(AGG_AlignRight ,AGG_AlignCenter );
    VG.Text(250.0 ,400.0 ,'Right-Center' ,true ,0 ,0 );

    VG.TextAlignment(AGG_AlignLeft ,AGG_AlignTop );
    VG.Text(250.0 ,450.0 ,'Left-Top' ,true ,0 ,0 );

    VG.TextAlignment(AGG_AlignCenter ,AGG_AlignTop );
    VG.Text(250.0 ,500.0 ,'Center-Top' ,true ,0 ,0 );

    VG.TextAlignment(AGG_AlignRight ,AGG_AlignTop );
    VG.Text(250.0 ,550.0 ,'Right-Top' ,true ,0 ,0 );

   // Gradients (Aqua Buttons)
   //=======================================
    VG.Font(FONT_VERDA ,20.0 ,false ,false ,AGG_VectorFontCache );

    xb1:=400;
    yb1:=80;
    xb2:=xb1 + 150;
    yb2:=yb1 + 36;

    clr.Construct (0 ,50 ,180 ,180 );
    VG.FillColor  (clr );
    clr.Construct (0 ,0 ,80 ,255 );
    VG.LineColor  (clr );
    VG.LineWidth  (1.0);
    VG.RoundedRect(xb1 ,yb1 ,xb2 ,yb2 ,12 ,18 );

    clr.Construct(0 ,0 ,0 ,0 );
    VG.LineColor (clr );

    c1.Construct         (100 ,200 ,255 ,255 );
    c2.Construct         (255 ,255 ,255 ,0 );
    VG.FillLinearGradient(xb1 ,yb1 ,xb1 ,yb1 + 30 ,c1 ,c2 );
    VG.RoundedRect       (xb1 + 3 ,yb1 + 2.5 ,xb2 - 3 ,yb1 + 30 ,9 ,18 ,1 ,1 );

    clr.Construct(0 ,0 ,50 ,200 );
    VG.FillColor (clr );
    VG.NoLine;
    VG.TextAlignment(AGG_AlignCenter ,AGG_AlignCenter );
    VG.Text(
     (xb1 + xb2 ) / 2.0 ,
     (yb1 + yb2 ) / 2.0 ,
     'Aqua Button' ,
     true ,0.0 ,0.0 );

    c1.Construct(0   ,0   ,255 ,0 );
    c2.Construct(100 ,255 ,255 ,255 );
    VG.FillLinearGradient(xb1 ,yb2 - 20 ,xb1 ,yb2 - 3 ,c1 ,c2 );
    VG.RoundedRect       (xb1 + 3 ,yb2 - 20 ,xb2 - 3 ,yb2 - 2 ,1 ,1 ,9 ,18 );

   // Aqua Button Pressed
    xb1:=400;
    yb1:=30;
    xb2:=xb1 + 150;
    yb2:=yb1 + 36;

    clr.Construct (0 ,50 ,180 ,180 );
    VG.FillColor  (clr );
    clr.Construct (0 ,0 ,0 ,255 );
    VG.LineColor  (clr );
    VG.LineWidth  (2.0 );
    VG.RoundedRect(xb1 ,yb1 ,xb2 ,yb2 ,12 ,18 );

    clr.Construct(0 ,0 ,0 ,0 );
    VG.LineColor (clr );

    c1.Construct         (60  ,160 ,255 ,255 );
    c2.Construct         (100 ,255 ,255 ,0 );
    VG.FillLinearGradient(xb1 ,yb1 + 2 ,xb1 ,yb1 + 25 ,c1 ,c2 );
    VG.RoundedRect       (xb1 + 3 ,yb1 + 2.5 ,xb2 - 3 ,yb1 + 30 ,9 ,18 ,1 ,1 );

    clr.Construct(0 ,0 ,50 ,255 );
    VG.FillColor (clr );
    VG.NoLine;
    VG.TextAlignment(AGG_AlignCenter ,AGG_AlignCenter );
    VG.Text(
     (xb1 + xb2 ) / 2.0 ,
     (yb1 + yb2 ) / 2.0 ,
     'Aqua Pressed' ,
     false ,0.0 );

    c1.Construct         (0 ,180 ,255 ,0 );
    c2.Construct         (0 ,200 ,255 ,255 );
    VG.FillLinearGradient(xb1 ,yb2 - 25 ,xb1 ,yb2 - 5 ,c1 ,c2 );
    VG.RoundedRect       (xb1 + 3 ,yb2 - 25 ,xb2 - 3 ,yb2 - 2 ,1 ,1 ,9 ,18 );

   // Basic Shapes -- Ellipse
   //===========================================
    VG.LineWidth(3.5 );
    VG.LineColor(20  ,80  ,80 );
    VG.FillColor(200 ,255 ,80 ,200 );
    VG.Ellipse  (450 ,200 ,50 ,90 );

   // Paths
   //===========================================
    VG.ResetPath;
    VG.FillColor (255 ,0 ,0   ,100 );
    VG.LineColor (0   ,0 ,255 ,100 );
    VG.LineWidth (2 );
    VG.MoveTo    (300 / 2 ,200 / 2 );
    VG.HorLineRel(-150 / 2 );
    VG.ArcRel    (150 / 2 ,150 / 2 ,0 ,true ,false ,150 / 2 ,-150 / 2 );
    VG.ClosePolygon;
    VG.DrawPath;

    VG.ResetPath;
    VG.FillColor (255 ,255 ,0   ,100 );
    VG.LineColor (0   ,0   ,255 ,100 );
    VG.LineWidth (2 );
    VG.MoveTo    (275 / 2 ,175 / 2 );
    VG.VerLineRel(-150 / 2 );
    VG.ArcRel    (150 / 2 ,150 / 2 ,0 ,false ,false ,-150 / 2 ,150 / 2 );
    VG.ClosePolygon;
    VG.DrawPath;

    VG.ResetPath;
    VG.NoFill;
    VG.LineColor(127 ,0 ,0 );
    VG.LineWidth(5 );
    VG.MoveTo   (600 / 2 ,350 / 2 );
    VG.LineRel  (50 / 2  ,-25 / 2 );
    VG.ArcRel   (25 / 2  ,25 / 2 ,deg2Rad(-30 ) ,false ,true ,50 / 2 ,-25 / 2 );
    VG.LineRel  (50 / 2  ,-25 / 2);
    VG.ArcRel   (25 / 2  ,50 / 2 ,deg2Rad(-30 ) ,false ,true ,50 / 2 ,-25 / 2 );
    VG.LineRel  (50 / 2  ,-25 / 2 );
    VG.ArcRel   (25 / 2  ,75 / 2 ,deg2Rad(-30 ) ,false ,true ,50 / 2 ,-25 / 2 );
    VG.LineRel  (50 ,-25 );
    VG.ArcRel   (25 / 2 ,100 / 2 ,deg2Rad(-30 ) ,false ,true ,50 / 2 ,-25 / 2 );
    VG.LineRel  (50 / 2 ,-25 / 2 );
    VG.DrawPath;

   // Master Alpha. From now on everything will be translucent
   //===========================================
    VG.MasterAlpha(0.85 );

   // Image Transformations
   //===========================================
    VG.ImageFilter  (AGG_Bilinear );
    VG.ImageResample(AGG_ResampleOnZoomOut );

   // Just copy image to the drawing surface
   // (No transformations apply)
   //-----------------
   { VG.CopyImage(Image2.Picture.Bitmap ,450 ,200 ); {}

   // Transform the whole image to the destination rectangle
   //-----------------
   { VG.TransformImage(Image2.Picture.Bitmap ,450 ,200 ,595 ,350 ); {}

   // Transform the rectangular part of the image to the destination rectangle
   //-----------------
   { VG.TransformImage(
     Image2.Picture.Bitmap ,60 ,60 ,
     Image2.Picture.Bitmap.Width - 60 ,
     Image2.Picture.Bitmap.Height - 60 ,
     450 ,200 ,595 ,350 ); {}

   // Transform the whole image to the destination parallelogram
   //-----------------
   { parl[0 ]:=450;
    parl[1 ]:=200;
    parl[2 ]:=595;
    parl[3 ]:=220;
    parl[4 ]:=575;
    parl[5 ]:=350;

    VG.TransformImage(Image2.Picture.Bitmap ,@parl[0 ] ); {}

   // Transform the rectangular part of the image to the destination parallelogram
   //-----------------
   { parl[0 ]:=450;
    parl[1 ]:=200;
    parl[2 ]:=595;
    parl[3 ]:=220;
    parl[4 ]:=575;
    parl[5 ]:=350;

    VG.TransformImage(
     Image2.Picture.Bitmap ,60 ,60 ,
     Image2.Picture.Bitmap.Width - 60 ,
     Image2.Picture.Bitmap.Height - 60 ,
     @parl[0 ] ); {}

   // Transform image to the destination path. The scale is determined by a rectangle
   //-----------------
   { VG.ResetPath;
    VG.MoveTo            (450 ,200 );
    VG.CubicCurveTo      (595 ,220 ,575 ,350 ,595 ,350 );
    VG.LineTo            (470 ,340 );
    VG.TransformImagePath(Image2.Picture.Bitmap ,450 ,200 ,595 ,350 ); {}

   // Transform image to the destination path.
   // The scale is determined by a rectangle
   //-----------------
    VG.ResetPath;
    VG.MoveTo      (450 ,200 );
    VG.CubicCurveTo(595 ,220 ,575 ,350 ,595 ,350 );
    VG.LineTo      (470 ,340 );
    VG.TransformImagePath(
     Image2.Picture.Bitmap ,60 ,60 ,
     Image2.Picture.Bitmap.Width - 60 ,
     Image2.Picture.Bitmap.Height - 60 ,
     450 ,200 ,595 ,350 ); {}

   // Add/Sub/Contrast Blending Modes
    VG.NoLine;
    VG.FillColor(70 ,70 ,0 );
    VG.BlendMode(AGG_BlendAdd );
    VG.Ellipse  (500 ,280 ,20 ,40 );

    VG.FillColor(255 ,255 ,255 );
    VG.BlendMode(AGG_BlendContrast );
    VG.Ellipse  (500 + 40 ,280 ,20 ,40 );

   // Radial gradient.
    VG.BlendMode         (AGG_BlendAlpha );
    c1.Construct         (255 ,255 ,0 , 0);
    c2.Construct         (0   ,0   ,127 );
    c3.Construct         (0   ,255 ,0 , 0);
    VG.FillRadialGradient(400 ,500 ,40 ,c1 ,c2 ,c3 );
    VG.Ellipse           (400 ,500 ,40 ,40 );

   // Gradient Text
   //-----------------
    VG.MasterAlpha(1 );

    c1.Construct(255 ,0 ,0 );
    c2.Construct(0 ,255 ,0 );
    c3.Construct(0 ,0 ,255 );

    VG.LineRadialGradient(165 ,590 ,430 ,c1 ,c2 ,c3 );
    VG.LineWidth(2 );
    VG.FillRadialGradient(165 ,590 ,430 ,c1 ,c2 ,c3 );

    VG.Font(FONT_VERDA ,29 ,false ,false ,AGG_VectorFontCache );

    VG.TextAlignment(AGG_AlignLeft ,AGG_AlignTop );
    VG.Text(165 ,590 ,'TAgg2D: Delphi API for AGG' );

   end;

end;

procedure TForm1.TrackBar1Change(Sender: TObject);
begin
 FormResize(Sender );
 Repaint;

end;

INITIALIZATION
 if Agg2DUsesFreeType then
  begin
   FONT_TIMES:='times.ttf';
   FONT_ARIAL:='arial.ttf';
   FONT_VERDA:='verdana.ttf';

  end
 else
  begin
   FONT_TIMES:='Times New Roman';
   FONT_ARIAL:='Arial';
   FONT_VERDA:='Verdana';

  end;

end.
 